package speiger.src.coreengine.rendering.input.window;

import java.util.List;

import org.lwjgl.glfw.GLFW;
import org.lwjgl.opengl.GL;
import org.lwjgl.opengl.GLCapabilities;
import org.lwjgl.system.CallbackI;

import speiger.src.collections.objects.lists.ObjectArrayList;
import speiger.src.coreengine.rendering.input.window.WindowCallback.ReloadFunction;
import speiger.src.coreengine.rendering.input.window.WindowManager.WindowBuilder;
import speiger.src.coreengine.utils.collections.FlagHolder;

public class Window {
	static final int FLAG_VISIBLE = 1;
	static final int FLAG_VSYNC = 2;
	static final int FLAG_FOCUS = 4;
	static final int FLAG_CPU_FPS_CAP = 8;
	static final int FLAG_FULL_SCREEN = 16;
	static final int FLAG_MAXIMIZED = 32;
	static final int FLAG_BORDERLESS = 64;
	static final int FLAG_CLOSE = 128;
	static final int FLAG_WINDOW_CHANGE = 256;
	static final int FLAG_FLOATING = 512;
	WindowManager manager;
	FlagHolder flags = new FlagHolder();
	long id;
	VideoMode mode;
	String title = "";
	int x;
	int y;
	int width;
	int height;
	
	int frameWidth;
	int frameHeight;
	
	int windowX;
	int windowY;
	int windowWidth;
	int windowHeight;
	
	final int antialiasing;
	List<WindowCallback> callbacks = new ObjectArrayList<>();
	GLCapabilities capabilities;
	
	protected Window(WindowBuilder builder) {
		manager = builder.manager;
		title = builder.title;
		width = builder.width;
		height = builder.height;
		antialiasing = builder.antiAlis;
		mode = builder.mode;
		flags.setFlag(FLAG_BORDERLESS, builder.borderless);
		flags.setFlag(FLAG_FULL_SCREEN, builder.fullScreen);
		flags.setFlag(FLAG_FLOATING, builder.floating);
		flags.setFlag(FLAG_VSYNC, builder.vsync);
		flags.setFlag(FLAG_CPU_FPS_CAP, builder.fpsCap);
		createDefaultWindowHints();
		GLFW.glfwWindowHint(GLFW.GLFW_SAMPLES, antialiasing);
		GLFW.glfwWindowHint(GLFW.GLFW_DECORATED, flags.isFlagNotSet(FLAG_FULL_SCREEN) && flags.isFlagSet(FLAG_BORDERLESS) ? 0 : 1);
		GLFW.glfwWindowHint(GLFW.GLFW_FLOATING, flags.isFlagNotSet(FLAG_FULL_SCREEN) && flags.isFlagSet(FLAG_FLOATING) ? 1 : 0);
		GLFW.glfwWindowHint(GLFW.GLFW_MAXIMIZED, flags.isFlagNotSet(FLAG_FULL_SCREEN) && flags.isFlagSet(FLAG_MAXIMIZED) ? 1 : 0);
		if(mode == null || mode.monitor() == null) throw new IllegalStateException("Monitor or Video Mode is missing: "+mode);
		Monitor monitor = mode.monitor();
		boolean fullscreen = builder.fullScreen;
		id = GLFW.glfwCreateWindow(fullscreen ? mode.width() : width, fullscreen ? mode.height() : height, title, builder.fullScreen ? monitor.id() : 0, manager.getPrimaryWindow());
		if(id == 0) throw new IllegalStateException("Window Couldn't be Created");
		manager.addWindow(this);
		createWindowListeners();
		GLFW.glfwMakeContextCurrent(id);
		capabilities = GL.createCapabilities(true);
		x = monitor.xOffset() + (builder.center ? (mode.width() / 2) - (width / 2) : 0);
		y = monitor.yOffset() + (builder.center ? (mode.height() / 2) - (height / 2) : 0);
		if(!fullscreen) GLFW.glfwSetWindowPos(id, x, y);
		GLFW.glfwSwapInterval(flags.isFlagSet(FLAG_VSYNC) ? 1 : 0);
		fetchWindowBounds();
	}
	
	protected void createDefaultWindowHints() {
		GLFW.glfwDefaultWindowHints();
		GLFW.glfwWindowHint(GLFW.GLFW_VISIBLE, GLFW.GLFW_FALSE);
		GLFW.glfwWindowHint(GLFW.GLFW_RESIZABLE, GLFW.GLFW_TRUE);
		GLFW.glfwWindowHint(GLFW.GLFW_CONTEXT_VERSION_MAJOR, 4);
		GLFW.glfwWindowHint(GLFW.GLFW_CONTEXT_VERSION_MINOR, 0);
		GLFW.glfwWindowHint(GLFW.GLFW_OPENGL_PROFILE, GLFW.GLFW_OPENGL_CORE_PROFILE);
		GLFW.glfwWindowHint(GLFW.GLFW_OPENGL_FORWARD_COMPAT, GLFW.GLFW_TRUE);
	}
	
	protected void createWindowListeners() {
		
	}
	
	protected void fetchWindowBounds() {
		int[] width = new int[1];
		int[] height = new int[1];
		GLFW.glfwGetWindowSize(id, width, height);
		this.frameWidth = width[0];
		this.frameHeight = height[0];
	}
	
	@SuppressWarnings("unchecked")
	public <T extends CallbackI> void addCallback(T listener, ReloadFunction<T> function) {
		WindowCallback callback = new WindowCallback(listener, (ReloadFunction<CallbackI>)function);
		callbacks.add(callback);
		callback.load(id);
	}
	
	public void title(String name) {
		if(name == null || title.equals(name)) return;
		title = name;
		GLFW.glfwSetWindowTitle(id, title);
	}
	
	public void vsync(boolean vsync) {
		if(!flags.setFlag(FLAG_VSYNC, vsync)) return;
		GLFW.glfwSwapInterval(vsync ? 1 : 0);
	}
	
	public void fpsCap(boolean fpsCap) {
		flags.setFlag(FLAG_CPU_FPS_CAP, fpsCap);
	}
	
	public void visible(boolean visible) {
		if(!flags.setFlag(FLAG_VISIBLE, visible)) return;
		if(visible) GLFW.glfwShowWindow(id);
		else GLFW.glfwHideWindow(id);
	}
	
	public void floating(boolean floating) {
		if(flags.isFlagNotSet(FLAG_FULL_SCREEN) && flags.setFlag(FLAG_FLOATING, floating)) {
			GLFW.glfwSetWindowAttrib(id, GLFW.GLFW_FLOATING, floating ? GLFW.GLFW_TRUE : GLFW.GLFW_FALSE);
		}
	}
	
	public void maximized(boolean maximized) {
		if(flags.isFlagNotSet(FLAG_FULL_SCREEN) && flags.setFlag(FLAG_MAXIMIZED, maximized)) {
			if(maximized) {
				windowX = x;
				windowY = y;
				windowWidth = width;
				windowHeight = height;
				GLFW.glfwMaximizeWindow(id);
			}
			else {
				x = windowX;
				y = windowY;
				width = windowWidth;
				height = windowHeight;
				GLFW.glfwRestoreWindow(id);
			}
			fetchWindowBounds();
		}
	}
	
	public void resizeable(boolean resizeable) {
		
	}
	
	public void borderless(boolean borderless) {
		
	}
	
	public void fullscreen(boolean fullscreen) {
		
	}
	
	public boolean position(int x, int y) {
		return false;
	}
	
	public boolean size(int width, int height) {
		return false;
	}
	
	public boolean width(int width) {
		return false;
	}
	
	public boolean height(int height) {
		return false;
	}
	
	public long id() { return id; }
	public VideoMode mode() { return mode; }
	
	public int x() { return x; }
	public int y() { return y; }
	public int width() { return frameWidth; }
	public int height() { return frameHeight; }
	public int screenWidth() { return width; }
	public int screenHeight() { return height; }
	public boolean changed() { return flags.isFlagSet(FLAG_WINDOW_CHANGE); }
	
	public String title() { return title; }
	public boolean isVsync() { return false; }
	public boolean isFPSCapped() { return false; }
	public boolean isVisible() { return false; }
	public boolean isFloating() { return false; }
	public boolean isMaximized() { return false; }
	public boolean isResizeable() { return false; }
	public boolean isBorderless() { return false; }
	public boolean isFullscreen() { return false; }
	public int antialiasing() { return antialiasing; }
}
